# Hybrid-PassCoDe
Extend PassCoDe parallel framework to a distributed framework on a hybrid system


# Prerequisites
1. Linux system and python, GSL(GNU Scientific Library) is required
  To obtain python and GSL, you can use apt-get, yum, or similar package managements to install, make sure the dev version are installed. Or you can build from source code, please refer to https://www.gnu.org/software/gsl/ and https://www.python.org

2. gcc, g++
  The gcc and g++ can also be installed via apt-get, yum, or similar package mangements. Please refer to https://gcc.gnu.org for more information.

3. openmpi or mpich
  To build from source code, please refer to https://www.open-mpi.org or https://www.mpich.org for more information. Otherwise you can use apt-get, yum, or similar package managements to install. Make sure the dev version are installed, too.

4. boost_mpi
  Please refer to http://www.boost.org for more information of boost library.
  To download and install boost lib, refer to http://www.boost.org/doc/libs/1_64_0/more/getting_started/unix-variants.html
  Please make sure boost is built with MPI support by using mpi lib, refer to http://www.boost.org/doc/libs/1_61_0/doc/html/mpi/getting_started.html#mpi.config for config and install of boost_mpi


# Dataset
1. The dataset must be in LIBSVM format. For details of LIBSVM data format, please refer to https://www.csie.ntu.edu.tw/~cjlin/libsvm/

2. To download the test dataset, please go to https://www.csie.ntu.edu.tw/~cjlin/libsvmtools/datasets/for classification or regression dataset.

3. To prepare dataset, after downloading the LIBSVM data,
$ cd /path/to/hdca/prep_data_script/orig
$ cp /path/to/dataname .
$ python div-train-test.py orig/
$ python prepare-data.py dataname
This will generate partitioned data in prep_data_script/cbin/


# Build
Use default makefile to make:
$ cd /path/to/hdca/
$ make
This will generate executable hdca in /path/to/hdca/


# Run
$ ./hdca --help
will print help information, regarding all parameter settings as below:
Usage: ./hdca [OPTIONS]
	-A <infile>  Prefix of the svm input file(s), actual filename must follow .<#nodes>.<node-id>
	-B <int>     Number of times subproblem is solved in a bulk iteration
	-f <int>     Loss function: 0 for hinge
	-H <int>     Number of local iterations by subproblem solver
	-I <int>     Number of bulk iterations to be executed
	-l <double>  Parameter lambda
	-M <int>     Subproblem solver: 0 for SDCA, 1 for PassCoDe
	-o <outfile> Output file : (default="hdca_run.txt")
	-p <int>     Number of (parts of) input files : (default=16)
	-S <int>     Size of update set : (default=num-nodes)
	-T <int>     Delay bound (tau) : (default="1")
	-t <int>     Number of threads
	-W <int>     Framework: 0 for CoCoA(+), 1 for Hybrid-DCA, 2 for Baseline

Note that CoCoA(+) framwork is too large and only for performance comparison, thus only header file provided, for running CoCoA(+), you need to directly download from https://github.com/gingsmith/cocoa. As well please refer to the paper 'Ma, C., Smith, V., Jaggi, M., Jordan, M. I., Richtarik, P., & Takac, M. Adding vs. Averaging in Distributed Primal-Dual Optimization. ICML 2015 - International Conference on Machine Learning.' for details of CoCoA and CoCoA(+) framework.

A sample command for running:
$ /path/to/hdca/hdca -W 1 -M 1 -f 0 -H 10000 -I 20 -B 5 -t 1 -S 4 -T 1 -p 4 -A /path/to/hdca/prep_data_script/cbin/dataname.train


# Misc:
1. In src/dataparser/instance_loader.hpp, function helpLoadData(): to load svm data correctly, for regression we need to set scanf of %lf and read corresponding double value, for classification, we set it to %i and read integer value

2. There are some sample script for different test cases that you can find in /path/to/hdca/test/

3. To run in UConn Hornet Cluster,
$ module load gcc/4.8.2
$ module load gsl/2.1
$ module load boost/1.58.0-gcc-mpi
$ module load mpi/openmpi/1.6.5-gcc
$ srun --mpi=openmpi--ntasks-per-node=4 -N 1 /path/to/hdca/hdca -W 1 -M 1 -f 0 -H 10000 -I 20 -B 5 -t 1 -S 4 -T 1 -p 4 -A /path/to/hdca/prep_data_script/cbin/dataname.train

4. The data is read as csr data format, while original data in file is LIBSVM format
 * original LIBSVM data example
 *
 * -1 4:1 6:1 18:1 21:1 35:1 40:1 57:1 63:1 67:1 73:1 74:1 77:1 80:1 83:1 
 * +1 5:1 6:1 15:1 24:1 38:1 40:1 51:1 61:1 71:1 72:1 74:1 76:1 80:1 83:1
 *
 * then after reading to problemData class, following values are set
 *
 * b = -1 1
 * m = 83
 * n = 2
 * row_ptr = 0 14
 * col_id = 3 5 17 20 34 39 56 62 66 72 73 76 79 82    4 5 14 23 37 39 ...etc...

