#ifndef __COCOA_NODE_HPP__
#define __COCOA_NODE_HPP__

#include "node.hpp"

template<typename L, typename D>
class CoCoA : public Node<L,D> {
  public:
    CoCoA(boost::mpi::communicator& world, Parameters &par, ProblemData<L,D> &ins, Solver<L,D> *s, LossFunction<L,D> *lf, StatsLogger<L,D> &sl) : Node<L,D>(world, par, ins, s, lf, sl) { }
    virtual ~CoCoA() {
    }
    virtual void run() {
      
      double start = 0;
      double finish = 0;
      double elapsedTime = 0;
      unsigned long communication_count = 0;

      cout << "communicate count = " << params.bulk_iters << endl;
      cout << "bulk_iteration count = " << params.bulk_size << endl;
      cout << "per thread iterations count = " << params.local_iters << endl;

      for (unsigned int t = 0; t < params.bulk_iters; t++) {

        start = gettime_();

        for (int jj = 0; jj < params.bulk_size; jj++, communication_count++) {
          cblas_set_to_zero(deltaW);
          solver->solve(w, deltaW);
          vall_reduce(world, deltaW, wBuffer);
          cblas_sum_of_vectors(w, wBuffer, params.gamma);
        }
        finish = gettime_();
        elapsedTime += finish - start;

        double primalError;
        double dualError;
        lf->computeObjectiveValue(w, dualError, primalError);

        double finalPrimalError;
        double finalDualError;
        vall_reduce(world, &primalError, &finalPrimalError, 1);
        vall_reduce(world, &dualError, &finalDualError, 1);

        if (params.verbose && (world.rank() == 0)) {
          stats_logger.writeTempLog(instance, communication_count, elapsedTime, w, finalDualError, finalPrimalError);
        }
      }

    }

  private:
    using Node<L,D>::world;
    using Node<L,D>::params;
    using Node<L,D>::instance;
    using Node<L,D>::stats_logger;
    using Node<L,D>::wBuffer;
    using Node<L,D>::deltaW;
    using Node<L,D>::w;
    using Node<L,D>::solver;
    using Node<L,D>::lf;
};


#endif // __COCOA_NODE_HPP__

