#ifndef __PROCESSING_NODE_HPP__
#define __PROCESSING_NODE_HPP__

#include "../params.hpp"
#include "../dataparser/structures.hpp"
#include "../dataparser/instance_loader.hpp"
#include "../utils/distributed_essentials.hpp"
#include "../utils/my_cblas_wrapper.hpp"
#include "../utils/utils.hpp"
#include "../utils/stats_logger.hpp"
#include "../lossfunction/hinge_loss.hpp"
#include "../solver/sdca.hpp"
#include "../solver/passcode.hpp"

template<typename L, typename D>
class Node {
  public:
    Node(boost::mpi::communicator& _world, Parameters &par, ProblemData<L,D> &ins, Solver<L,D> *s, LossFunction<L,D> *_lf, StatsLogger<L,D> &sl) : world(_world), params(par), instance(ins), solver(s), lf(_lf), stats_logger(sl) {
      init();
    }
    virtual ~Node() {
    }
    virtual void run() = 0;

  private:
    void init() {
      unsigned int finalM;
      vall_reduce_maximum(world, &instance.m, &finalM, 1);
      cout << "Local m " << instance.m << "   global m " << finalM << endl; 
      instance.m = finalM;
      instance.lambda = params.lambda;

      wBuffer.resize(instance.m + 1);
      deltaW.resize(instance.m + 1); // piggyback local dual objective
      w.resize(instance.m+1);
      instance.x.resize(instance.n);
      cout << "x.size=" << instance.x.size() << endl;
      cblas_set_to_zero(instance.x);

      vall_reduce(world, &instance.n, &instance.total_n, 1);
      cout << "Local n " << instance.n << "   global n " << instance.total_n << endl;

      cblas_set_to_zero(w);
      cblas_set_to_zero(deltaW);
      vall_reduce(world, deltaW, w);

      instance.oneOverLambdaN = 1 / (0.0 + instance.total_n * instance.lambda);

      if (params.APPROX) {
        params.gamma = 1;
        instance.penalty = params.update_set + 0.0;
      } else {
        params.gamma = 1 / (params.update_set + 0.0);
        instance.penalty = 1;
      }

    }

  protected:
    boost::mpi::communicator& world;
    Parameters& params;
    ProblemData<L,D> &instance;
    Solver<L,D> *solver;
    LossFunction<L,D> *lf;
    StatsLogger<L,D> &stats_logger;
    std::vector<D> wBuffer;
    std::vector<D> deltaW;
    std::vector<D> w;
};


#endif // __PROCESSING_NODE_HPP__

