#ifndef __WORKER_HPP__
#define __WORKER_HPP__

template<typename L, typename D>
class Worker : public Node<L,D> {
  public:
    Worker(boost::mpi::communicator& world, Parameters &par, ProblemData<L,D> &ins, Solver<L,D> *s, LossFunction<L,D> *_lf, StatsLogger<L,D> &sl) : Node<L,D>(world, par, ins, s, _lf, sl) { }
    ~Worker() {}
    void run() {
      size_t total_rounds = params.bulk_iters * params.bulk_size;
      size_t num_rounds = 0;
      while (1) {
        cblas_set_to_zero(deltaW);
        solver->solve(w, deltaW);
        lf->computeDualObjective(deltaW[instance.m]); 
        world.send(0, 16, deltaW);
        if (++num_rounds >= total_rounds) break;
        world.recv(0, 16, w);
      }
    }
  private:
    using Node<L,D>::world;
    using Node<L,D>::params;
    using Node<L,D>::deltaW;
    using Node<L,D>::w;
    using Node<L,D>::solver;
    using Node<L,D>::instance;
    using Node<L,D>::lf;
};

#endif // __WORKER_HPP__
