#ifndef __HINGE_LOSS_HPP__
#define __HINGE_LOSS_HPP__

#include "loss_function.hpp"

template<typename L, typename D>
class HingeLoss : public LossFunction<L, D> {
public:
	HingeLoss(ProblemData<L,D>& pd, Parameters &par) : LossFunction<L,D>(pd, par) {
    resetInstance();
  }
	virtual ~HingeLoss() {}

  virtual void resetInstance() {
		instance.Li.resize(instance.n);
		instance.vi.resize(instance.n);
    cout << "instance.n=" << instance.n << endl;
    cout << "instance.penalty=" << instance.penalty << endl;
    cout << "instance.oneoverN=" << instance.oneOverLambdaN << endl;
		for (L idx = 0; idx < instance.n; idx++) {
			D norm = cblas_l2_norm(
					instance.A_csr_row_ptr[idx + 1]
										   - instance.A_csr_row_ptr[idx],
										   &instance.A_csr_values[instance.A_csr_row_ptr[idx]], 1);
			instance.Li[idx] = 1 / (norm * norm * instance.penalty * instance.oneOverLambdaN);

			instance.vi[idx] = norm * norm;
		}
  }

	virtual void computeDualObjective(D &localError){
		localError = 0;
		for (unsigned int i = 0; i < instance.n; i++) {
			D tmp = -instance.x[i];
			localError += tmp;
		}
	}

	virtual void computePrimalObjective(ProblemData<L,D> &generic_instance, std::vector<D> & w, D &localHingeLoss){
		localHingeLoss = 0;
		for (unsigned int idx = 0; idx < generic_instance.n; idx++) {
			D dotProduct = 0;
			for (L i = generic_instance.A_csr_row_ptr[idx];
					i < generic_instance.A_csr_row_ptr[idx + 1]; i++) {
				dotProduct += (w[generic_instance.A_csr_col_idx[i]])
									* generic_instance.A_csr_values[i];
			}
			D tmp = 1 - generic_instance.b[idx] * dotProduct;
			if (tmp > 0) {
				localHingeLoss += tmp;
			}
		}
	}

	virtual void computeObjectiveValue(std::vector<D> & w, D &localError, D &localHingeLoss){
        computeDualObjective(localError);
        computePrimalObjective(instance, w, localHingeLoss);
	}

private:
    using LossFunction<L,D>::instance;


};

#endif /* __HINGE_LOSS_HPP__ */
