#ifndef __HDCA_PARAMS_H__
#define __HDCA_PARAMS_H__

#include <string>
#include <fstream>

const string default_out = "hdca_run.txt";

enum SolverID : unsigned char {
  SDCA, PASSCODE
};

enum LossFunctionID : unsigned char {
  HINGE, SQUAREDHINGE, LOGISTICS
};

enum FrameworkID : unsigned char {
  COCOA, HDCA, BASELINE
};


struct Parameters {
  double lambda;
  size_t num_threads;
  size_t local_iters;
  size_t bulk_iters;
  size_t bulk_size;
  bool APPROX;
  size_t update_set;
  size_t update_tau;
  LossFunctionID loss_function;
  SolverID sub_solver;
  std::string input_file_prefix;
  int input_num_parts;
  std::string output_file_name;
  std::ofstream out;
  bool verbose;
  double gamma;
  FrameworkID framework;
  Parameters() :
    lambda(0.0001),
    num_threads(4),
    local_iters(1000),
    bulk_iters(10),
    bulk_size(5),
    APPROX(true),
    update_set(0),
    update_tau(1),
    loss_function(LossFunctionID::HINGE),
    sub_solver(SolverID::SDCA),
    input_num_parts(16),
    output_file_name(default_out),
    verbose(true),
    gamma(1),
    framework(FrameworkID::COCOA)
  {}
};

#endif // __HDCA_PARAMS_H__
