#ifndef __SOLVER_PASSCODE_HPP__
#define __SOLVER_PASSCODE_HPP__

#include "solver.hpp"
#include "omp.h"

#include "../utils/my_cblas_wrapper.hpp"

template<typename L, typename D>
class PassCode : public Solver<L, D> {
public:
	PassCode(ProblemData<L,D>& pd, Parameters& params) : Solver<L,D>(pd, params) {
    deltaAlpha.resize(instance.n);
  }
	virtual ~PassCode() {}
  void solve(std::vector<D> &w, std::vector<D> &deltaW) {
    cblas_set_to_zero(deltaAlpha);
#pragma omp parallel
    {
      int tid = omp_get_thread_num();
      L len = (instance.n + params.num_threads - 1) / params.num_threads;
      L begin = tid*len, end = min((tid+1)*len, instance.n);
      L size = end - begin;
      for (unsigned int it = 0; it < params.local_iters; it++) {
        solveSingleIter(w, deltaW, begin, size);
      }
    }
    cblas_sum_of_vectors(instance.x, deltaAlpha, params.gamma);
  }
private:
  void solveSingleIter(std::vector<D> &w, std::vector<D> &deltaW, L start, L size) {
    L idx = start + (rand() / (0.0 + RAND_MAX)) * size;
    D dotProduct = 0;
    for (L i = instance.A_csr_row_ptr[idx]; i < instance.A_csr_row_ptr[idx + 1]; i++) {
      dotProduct += (w[instance.A_csr_col_idx[i]] + instance.penalty * deltaW[instance.A_csr_col_idx[i]])
        * instance.A_csr_values[i];
    }
    D alphaI = instance.x[idx] + deltaAlpha[idx];
    D deltaAl = 0;
    D part = (1.0 - instance.b[idx] * dotProduct) * instance.Li[idx];
    deltaAl = (part > 1 - alphaI) ? 1 - alphaI : (part < -alphaI ? -alphaI : part);
    deltaAlpha[idx] += deltaAl;
    for (L i = instance.A_csr_row_ptr[idx]; i < instance.A_csr_row_ptr[idx + 1]; i++) {
#pragma omp atomic
      deltaW[instance.A_csr_col_idx[i]] += instance.oneOverLambdaN
        * instance.A_csr_values[i] * deltaAl * instance.b[idx];
    }
  }
private:
  using Solver<L,D>::params;
  using Solver<L,D>::instance;
  std::vector<D> deltaAlpha;
};

#endif // __SOLVER_PASSCODE_HPP__

