#ifndef __SOLVER_SDCA_HPP__
#define __SOLVER_SDCA_HPP__

#include "solver.hpp"

#include "../utils/my_cblas_wrapper.hpp"

template<typename L, typename D>
class Sdca : public Solver<L, D> {
public:
	Sdca(ProblemData<L,D>& pd, Parameters& params) : Solver<L,D>(pd, params) {
    deltaAlpha.resize(instance.n);
  }
	virtual ~Sdca() {}
  void solve(std::vector<D> &w, std::vector<D> &deltaW) {
    cblas_set_to_zero(deltaAlpha);
    for (unsigned int it = 0; it < params.local_iters; it++) {
      solveSingleIter(w, deltaW);
    }
    cblas_sum_of_vectors(instance.x, deltaAlpha, params.gamma);
  }
private:
  void solveSingleIter(std::vector<D> &w, std::vector<D> &deltaW) {
    L idx = rand() / (0.0 + RAND_MAX) * instance.n;
          //cout << "idx=" << idx << endl;
    D dotProduct = 0;
    for (L i = instance.A_csr_row_ptr[idx]; i < instance.A_csr_row_ptr[idx + 1]; i++) {
      dotProduct += (w[instance.A_csr_col_idx[i]] + instance.penalty * deltaW[instance.A_csr_col_idx[i]])
        * instance.A_csr_values[i];
    }
          //cout << "dotProduct=" << dotProduct << endl;
    D alphaI = instance.x[idx] + deltaAlpha[idx];
          //cout << "x[idx]=" << instance.x[idx] << ", deltaAl[idx]=" << deltaAlpha[idx] << endl;
    D deltaAl = 0;
          //cout << "b[idx]=" << instance.b[idx] << ", Li[idx]=" << instance.Li[idx] << endl;
    D part = (1.0 - instance.b[idx] * dotProduct) * instance.Li[idx];
    deltaAl = (part > 1 - alphaI) ? 1 - alphaI : (part < -alphaI ? -alphaI : part);
          //cout << "deltaAl=" << deltaAl << endl;
    deltaAlpha[idx] += deltaAl;
    for (L i = instance.A_csr_row_ptr[idx]; i < instance.A_csr_row_ptr[idx + 1]; i++) {
      deltaW[instance.A_csr_col_idx[i]] += instance.oneOverLambdaN
        * instance.A_csr_values[i] * deltaAl * instance.b[idx];
    }
  }
private:
  using Solver<L,D>::params;
  using Solver<L,D>::instance;
  std::vector<D> deltaAlpha;
};

#endif // __SOLVER_SDCA_HPP__

